<?php

/**
 * @file
 *  The drush Issue Queue manager
 */



/**
 * Implementation of hook_drush_command().
 */
function iq_drush_command() {
  $items['iq-info'] = array(
    'description' => 'Show information about an issue from the queue on drupal.org.',
    'examples' => array(
      'drush iq-info 1234' => 'Get info on issue 1234.',
      'drush iq-info http://drupal.org/node/1234' => 'Get info on an issue identified by its URL.',
      'drush iq-info 1234-#3' => 'Get info on the patch on the third comment of issue 1234.',
      'drush iq-info 1234-5678' => 'Get info on the patch on comment id 5678 of issue 1234.',
    ),
    'arguments' => array(
      'number' => 'The issue number.',
    ),
    'required-arguments' => TRUE,
    'options' => array(
      'pipe' => 'Print the full issue info data structure.',
    ),
    'aliases' => array('iqi'),
    'bootstrap' => DRUSH_BOOTSTRAP_DRUSH,
    'topics' => array('docs-iq-commands'),
  );
  $items['iq-branch'] = array(
    'description' => 'List branches created by iq commands and switch to or delete them.',
    'examples' => array(
      'drush iq-branch' => 'Let the user select the iq branch to switch to from a list of available branches.',
      'drush iq-branch delete' => 'Prompt the user and then delete the selected branch',
      'drush iq-branch list' => 'Lists all iq branches.',
      'drush iq-branch 1234' => 'Switch back to the iq branch corresponding to issue 1234.  Will create a new working branch if one does not already exist.',
    ),
    'arguments' => array(
      'branch command' => 'The branch operation to perform.  One of switch, list or delete. Optional; defaults to switch.',
      'number' => 'The issue number.',
    ),
    'options' => array(
      'all' => 'Delete all branches.  Only applicable to the "iq-branch delete" command.',
      'base' => 'When creating a new branch while working in a detached HEAD state, creates a branch to serve as the base for future diffs and merges.  Only applicable to the "iq-branch 1234" (create) command.',
    ),
    'aliases' => array('iqb'),
    'bootstrap' => DRUSH_BOOTSTRAP_DRUSH,
    'topics' => array('docs-iq-commands'),
  );
  $items['iq-create-commit-comment'] = array(
    'description' => 'Create a commit comment for the specified issue number.',
    'examples' => array(
      'drush iq-create-commit-comment 1234' => 'Generate a commit comment using the title from issue 1234, and crediting every user who provided attachments.',
    ),
    'arguments' => array(
      'number' => 'The issue number.',
    ),
    'aliases' => array('iqccc', 'ccc'),
    'bootstrap' => DRUSH_BOOTSTRAP_DRUSH,
    'topics' => array('docs-iq-commands'),
  );
  $items['iq-apply-patch'] = array(
    'description' => 'Look up the most recent patch attached to the specified issue, and apply it to its project.',
    'examples' => array(
      'drush iq-apply-patch 1234' => 'Apply the newest patch attached to issue 1234.',
      'drush iq-apply-patch 1234-#5' => 'Apply the patch attached to the fifth comment of issue 1234.',
      'drush iq-apply-patch 1234-5678' => 'Apply the patch attached to comment id 5678 of issue 1234.',
      'drush iq-apply-patch 1234 --select' => 'Show all patches attached to issue 1234, and prompt for the one to apply.',
    ),
    'arguments' => array(
      'number' => 'The issue number. Optional; default is to attempt to find a patch from the issue open in a browser window.',
    ),
    'options' => array(
      'no-prefix' => 'Patch was created with --no-prefix, and therefore should be applied with -Np0 instead of -Np1.  Optional; default is to try both.',
      'no-git' => 'Do not execute any git commands. Default is to create a new branch for the issue and commit the patch.',
      'no-commit' => 'Create a new branch, but do not commit the patch to the issue working branch after applying it. Optional. Patches created by git format-patch are always committed.',
      'keep-patch' => 'Keep the patchfile after applying it.  Default is to delete the patchfile.',
      'dry-run' => 'Do not apply the patch; just print out messages about what would be done.',
      'select' => 'Prompt for which patch to apply.  Optional; default is newest patch.',
      'base' => 'When applying a patch while working in a detached HEAD state, creates a branch to serve as the base for future diffs and merges.',
    ),
    'aliases' => array('patch', 'iqa', 'am'),
    'bootstrap' => DRUSH_BOOTSTRAP_DRUSH,
    'topics' => array('docs-iq-commands'),
  );
  $items['iq-diff'] = array(
    'description' => 'Create a diff.  Uses `git format-patch`, so author information is included.',
    'examples' => array(
      'drush iq-diff' => 'Create a diff between the current branch and its upstream branch.',
      'drush iq-diff --commit' => 'Commit the changes to the current branch after creating the diff.',
    ),
    'arguments' => array(
      'number' => 'The issue number.',
    ),
    'options' => array(
      'no-prefix' => 'Create patch with no prefix.  Not recommended; patch will have to be applied with -Np0 instead of -Np1.',
      'no-git' => 'Do not execute any git commands; just run diff. Default is to use git format-patch.',
      'commit' => 'Commit change to current patch review branch. Optional; default is to leave changes unstaged.',
      'no-squash' => 'Show all commits in the branch, like a standard format-patch. Default is to squash into a single commit.',
      'rebase' => 'Rebase before creating patch.',
      'message' => 'Commit message.  Optional; default is to use `drush iq-create-commit-comment`.',
    ),
    'aliases' => array('diff', 'iqd'),
    'bootstrap' => DRUSH_BOOTSTRAP_DRUSH,
    'topics' => array('docs-iq-commands'),
  );
  $items['iq-reset'] = array(
    'description' => 'Stop working on a patch, and return to the original branch.',
    'examples' => array(
      'drush iq-reset' => 'Safely go back to the original branch. Work on the patch may resume later by returning to the working branch.',
      'drush iq-reset --hard' => 'Permanently delete all changes, and go back to the original branch.',
    ),
    'options' => array(
      'hard' => 'Also delete the working branch.',
    ),
    'aliases' => array('reset', 'iqr'),
    'bootstrap' => DRUSH_BOOTSTRAP_DRUSH,
    'topics' => array('docs-iq-commands'),
  );
  $items['iq-merge'] = array(
    'description' => 'Merge patch working branch into the original branch.',
    'examples' => array(
      'drush iq-merge' => 'Go back to the original branch, bringing all commits along.',
      'drush iq-merge --squash' => 'Merge multiple commits into a single commit when merging. May discard some author credit, as only the last commit is recorded.',
    ),
    'options' => array(
      'squash' => 'Merge all commits into one, keeping only the last.',
      'no-commit' => 'Skip committing unstaged changes.  Default is to commit everything.',
    ),
    'aliases' => array('merge', 'iqm'),
    'bootstrap' => DRUSH_BOOTSTRAP_DRUSH,
    'topics' => array('docs-iq-commands'),
  );
  $items['iq-lint'] = array(
    'description' => 'Run lint (php -l) on files that have changed in the current patch.',
    'examples' => array(
      'drush iq-lint' => 'Run lint on modified files.',
    ),
    'aliases' => array('lint', 'iql'),
    'bootstrap' => DRUSH_BOOTSTRAP_DRUSH,
    'topics' => array('docs-iq-commands'),
  );
  $items['docs-iq-commands'] = array(
    'description' => 'Issue queue commands for applying and creating patches from drupal.org.',
    'hidden' => TRUE,
    'topic' => TRUE,
    'bootstrap' => DRUSH_BOOTSTRAP_DRUSH,
    'callback' => 'drush_print_file',
    'callback arguments' => array(dirname(__FILE__) . '/docs/iq-commands.html'),
  );

  return $items;
}

function drush_iq_diff($number = NULL) {
  $branch_merges_with = FALSE;
  $issue_info = array();
  $remote = FALSE;
  if (isset($number)) {
    $issue_info = drush_iq_get_info($number);
    if (!$issue_info) {
      return FALSE;
    }
    $branch = _drush_iq_get_branch($issue_info);
    $dir = _drush_iq_project_dir($issue_info);
    if (!$dir) {
      return drush_set_error('DRUSH_IQ_DIFF_NO_PROJECT', dt('Could not find the project directory for !n', array('!n' => $number)));
    }
  }
  else {
    $dir = _drush_iq_git_root_dir_from_cwd();
    $branch = _drush_iq_get_branch_at_dir($dir);
    $branch_merges_with = _drush_iq_get_branch_merges_with($dir);
    if ($branch_merges_with === FALSE) {
      return FALSE;
    }
    if (!empty($branch_merges_with)) {
      $remote = _drush_iq_get_best_remote($branch_merges_with);
      $branch_merges_with = _drush_iq_merge_branch_with_remote($remote, $branch_merges_with);
    }
  }
  if (!$branch_merges_with && ($branch != "(no branch)")) {
    $branch_merges_with = 'master';
  }

  // Not under git revision control is an error
  if (empty($branch)) {
    return drush_set_error('DRUSH_NO_VCS', dt("Error: drush can only produce diffs of projects under git version control"));
  }

  // If the user did not provide an issue number, but we can find
  // that a git branch has been made at the current working directory,
  // then we will expect that the branch tag is in the format of
  // "arbitrary-prequel-ISSUENUMBER".  If it is in this form, drush_iq_get_info
  // will be able to find the issue number.
  if (isset($branch) && ($branch != "master") && empty($issue_info)) {
    $issue_info_id = drush_iq_issue_number($branch);
    if (!empty($issue_info_id)) {
      $number = $issue_info_id['id'];
      $issue_info = drush_iq_download_info($issue_info_id);
    }
  }
  // Add on extra flags
  $extra = "";
  if (drush_get_option('no-prefix', FALSE)) {
    $extra .= ' --no-prefix';
  }

  $cwd = getcwd();
  $committed = FALSE;
  $squashed = FALSE;
  drush_op('chdir', $dir);
  $patch_description = "drush-iq";
  if (!empty($issue_info)) {
    $comment_number = array_key_exists('comment-number', $issue_info) ? $issue_info['comment-number'] : 1 + count($issue_info['comments']);
    $patch_description = _drush_iq_make_description($issue_info['title']) . '-' . $issue_info['id'] . '-' . $comment_number . '.patch';
  }
  drush_log(dt("# Create patch: !patch_description (branch: !b merge: !m)", array('!patch_description' => $patch_description, '!b' => $branch, '!m' => $branch_merges_with)), 'notice');

  // If we have modified the master branch (not recommended), just run git diff
  if (($branch == "master") || ($branch == "(no branch)") || ($branch == $branch_merges_with) || drush_get_option('no-git', FALSE)) {
    $result = drush_shell_exec_interactive("git diff $extra %s", $branch_merges_with);
  }
  // Changes are being made under a local branch.  Do the recommended procedure for creating the diff.
  else {
    $commit_comment = _drush_iq_create_commit_comment($issue_info, drush_get_option('committer', TRUE), drush_get_option('message', FALSE));
    $result = drush_shell_exec("git status -s");
    $status_output = drush_shell_exec_output();
    // If there are unstaged changes, commit them
    if (!empty($status_output)) {
      // TODO: can we easily ignore *.patch, etc., or easily add only modified files?
      $result = drush_shell_exec("git add -u .");
      $result = drush_shell_exec("git commit -m %s", $commit_comment);
      if (!$result) {
        return drush_set_error('DRUSH_IQ_DIFF_CANNOT_COMMIT', dt("Commit failed."));
      }
      $committed = TRUE;
    }
    // If the branch to merge with is remote, then fetch and rebase
    if (drush_get_option('rebase', FALSE)) {
      if ($remote) {
        drush_shell_exec("git fetch %s %s && git rebase %s", $remote, $branch, $branch_merges_with);
      }
      else {
        drush_log(dt('--rebase ignored because !branch has no upstream branch.', array('!branch' => $branch)), 'warning');
      }
    }

    // Check to see if there are multiple commits on this branch. If there
    // are, make a new temporary branch and squash-merge all commits there.
    if (!drush_get_option('no-squash', FALSE)) {
      $result = drush_shell_exec("git log --oneline %s..HEAD", $branch_merges_with);
      $commits_output = drush_shell_exec_output();
      if (count($commits_output) > 1) {
        drush_shell_exec('git branch -D drush-iq-temp 2>/dev/null');
        drush_shell_exec('git checkout %s', $branch_merges_with);
        drush_shell_exec('git checkout -b drush-iq-temp');
        drush_shell_exec('git merge --squash %s', $branch);
        drush_shell_exec("git commit -m %s", $commit_comment);
        $squashed = TRUE;
      }
    }

    $result = drush_shell_exec_interactive("git format-patch %s -k $extra --stdout", $branch_merges_with);
    if ($squashed) {
      drush_shell_exec('git checkout %s', $branch);
      drush_shell_exec('git branch -D drush-iq-temp  2>/dev/null');
    }
    if ($committed && !drush_get_option('commit', FALSE)) {
      // Get rid of our commit
      drush_shell_exec("git reset HEAD~1");
    }
  }
  drush_op('chdir', $cwd);
  return $issue_info;
}

/**
 * iq-branch command callback
 */
function drush_iq_branch($command = 'switch', $number = NULL) {
  if (!isset($number)) {
    $issue_info_id = drush_iq_issue_number($command);
    if (!empty($issue_info_id)) {
      $number = $command;
      $command = 'switch';
    }
  }
  $branch_command_aliases = array(
    'l' => 'list',
    'ls' => 'list',
    'd' => 'delete',
    'del' => 'delete',
  );
  if (array_key_exists($command, $branch_command_aliases)) {
    $command = $branch_command_aliases[$command];
  }
  $function = 'drush_iq_branch_' . $command;
  if (function_exists($function)) {
    $issue_info = _drush_iq_find_project_dir($number);
    if ($issue_info == FALSE) {
      return FALSE;
    }
    return $function($issue_info);
  }
  else {
    return drush_set_error('DRUSH_IQ_UNKNOWN_BRANCH_COMMAND', dt('Unknown branch command "!command"', array('!command' => $command)));
  }
}

/**
 * iq-branch list command callback
 */
function _drush_iq_branch_list($issue_info) {
  $dir = _drush_iq_project_dir($issue_info);
  $result = drush_shell_cd_and_exec($dir, "git branch");
  $branch_output = drush_shell_exec_output();
  $iq_branches = array();
  foreach($branch_output as $line) {
    $branch_name = substr($line, 2);
    if (substr($branch_name, 0, 9) == 'drush-iq-') {
      $iq_branches[] = $branch_name;
    }
  }
  return $iq_branches;
}

function _drush_iq_branch_get_selections($issue_info) {
  $iq_branches = _drush_iq_branch_list($issue_info);
  $selections = array();
  if (array_key_exists('id', $issue_info)) {
    foreach ($iq_branches as $branch_name) {
      if (strstr('-' . $issue_info['id'] . '-', $branch_name) !== FALSE) {
        $selections[$branch_name] = $branch_name;
      }
    }
  }
  else {
    foreach ($iq_branches as $branch_name) {
      $selections[$branch_name] = $branch_name;
    }
  }
  return $selections;
}

function _drush_iq_branch_select($issue_info, $prompt) {
  $selections = _drush_iq_branch_get_selections($issue_info);
  if (empty($selections)) {
    return '';
  }
  $branch = drush_choice($selections, $prompt);
  if ($branch === FALSE) {
    return drush_user_abort();
  }
  return $branch;
}

/**
 * iq-branch list command callback
 */
function drush_iq_branch_list($issue_info) {
  $iq_branches = _drush_iq_branch_list($issue_info);
  drush_print(implode("\n", $iq_branches));
}

/**
 * iq-branch switch command callback
 */
function drush_iq_branch_switch($issue_info) {
  $branch = _drush_iq_branch_select($issue_info, dt('Select branch to switch to'));
  if ($branch === FALSE) {
    return FALSE;
  }
  if (empty($branch)) {
    if (array_key_exists('id', $issue_info)) {
      $confirm = drush_confirm(dt('No working branches available for issue !n.  Would you like to create one?', array('!n' => $issue_info['id'])));
      if (!$confirm) {
        return drush_user_abort();
      }
      return _drush_iq_create_branch($issue_info);
    }
    else {
      return drush_set_error('DRUSH_IQ_NO_BRANCH', dt('No working branches created by Drush iq commands available.'));
    }
  }
  $dir = _drush_iq_project_dir($issue_info);
  $result = drush_shell_cd_and_exec($dir, "git checkout %s", $branch);
  if ($result) {
    drush_log(dt('Switched to branch !branch', array('!branch' => $branch)), 'ok');
  }
  return $branch;
}

/**
 * iq-branch delete command callback
 */
function drush_iq_branch_delete($issue_info) {
  $dir = _drush_iq_project_dir($issue_info);
  if (drush_get_option('all', FALSE)) {
    $selections = _drush_iq_branch_get_selections($issue_info);
    drush_print("The following branches exist:\n");
    drush_print(implode("\n", $selections));
    $confirm = drush_confirm(dt('Would you like to delete them all?'));
    if (!$confirm) {
      return drush_user_abort();
    }
    foreach ($selections as $branch) {
      $result = drush_shell_cd_and_exec($dir, "git branch -D %s", $branch);
      if ($result) {
        drush_log(dt('Deleted branch !branch', array('!branch' => $branch)), 'ok');
      }
    }
  }
  $branch = _drush_iq_branch_select($issue_info, dt('Select branch to delete'));
  if ($branch === FALSE) {
    return FALSE;
  }
  if (empty($branch)) {
    if (array_key_exists('id', $issue_info)) {
      return drush_set_error('DRUSH_IQ_NO_BRANCH', dt('No working branches available for issue !n.', array('!n' => $issue_info['id'])));
    }
    else {
      return drush_set_error('DRUSH_IQ_NO_BRANCH', dt('No working branches created by Drush iq commands available.'));
    }
  }
  $result = drush_shell_cd_and_exec($dir, "git branch -D %s", $branch);
  if ($result) {
    drush_log(dt('Deleted branch !branch', array('!branch' => $branch)), 'ok');
  }
  return $branch;
}

/**
 * iq-info command callback
 */
function drush_iq_info($number = NULL) {
  if (isset($number)) {
    $issue_info = drush_iq_get_info($number);
  }
  else {
    $dir = _drush_iq_git_root_dir_from_cwd();
    $branch = _drush_iq_get_branch_at_dir($dir);
    $issue_info = drush_iq_get_info($branch);
  }
  if ($issue_info === FALSE) {
    return FALSE;
  }
  $issue_info = _drush_iq_select_patch($issue_info);

  if (drush_get_option('pipe')) {
    drush_print_pipe(var_export($issue_info, TRUE));
  }
  else {
    $label_map = _drush_iq_label_map();
    $rows = array();

    foreach($label_map as $label => $key) {
      if (array_key_exists($key, $issue_info)) {
        $rows[] = array($label, ':', $issue_info[$key]);
      }
    }

    drush_print_table($rows);
  }
  return $issue_info;
}

/**
 * iq-create-commit-comment command callback
 */
function drush_iq_create_commit_comment($number = NULL) {
  if (isset($number)) {
    $issue_info = drush_iq_get_info($number);
  }
  else {
    $dir = _drush_iq_git_root_dir_from_cwd();
    $branch = _drush_iq_get_branch_at_dir($dir);
    $issue_info = drush_iq_get_info($branch);
  }
  if ($issue_info === FALSE) {
    return FALSE;
  }

  $result = _drush_iq_create_commit_comment($issue_info, drush_get_option('committer', FALSE));
  drush_print($result);
  return $result;
}

/**
 * Find the project directory from the given issue number,
 * or from the cwd if no issue number is given.
 */
function _drush_iq_find_project_dir($number = NULL) {
  $issue_info = array();
  if (isset($number)) {
    $issue_info = drush_iq_get_info($number);
    if (!$issue_info) {
      return FALSE;
    }
    $dir = _drush_iq_project_dir($issue_info);
    if (!$dir) {
      return drush_set_error('DRUSH_IQ_NO_PROJECT', dt('Could not find the project directory for !n', array('!n' => $number)));
    }
  }
  else {
    $dir = _drush_iq_git_root_dir_from_cwd();
    $issue_info['project-dir'] = $dir;
  }
  return $issue_info;
}

/**
 * Abandon the current branch
 */
function drush_iq_reset($number = NULL) {
  $cwd = getcwd();
  $issue_info = _drush_iq_find_project_dir($number);
  if ($issue_info == FALSE) {
    return FALSE;
  }
  $dir = _drush_iq_project_dir($issue_info);
  $result = drush_shell_cd_and_exec($dir, "git status -s");
  $status_output = drush_shell_exec_output();
  $unstaged_changes = (!empty($status_output));
  $branch = _drush_iq_get_branch_at_dir($dir);
  $branch_merges_with = _drush_iq_get_branch_merges_with($dir);
  if ($branch_merges_with === FALSE) {
    return FALSE;
  }
  if (empty($branch_merges_with)) {
    $branch_merges_with = 'master';
  }
  $delete = drush_get_option('hard', FALSE);
  $issue_info['branch'] = $branch;
  $issue_info['mergeBranch'] = $branch_merges_with;

  $reset_message = dt("Would you like to reset to the branch !merges_with? ", array('!merges_with' => $branch_merges_with, '!branch' => $branch));

  if (($branch == '(no branch)') || ($branch_merges_with == $branch)) {
    $reset_message = "";
    if (!$unstaged_changes) {
      return drush_set_error('DRUSH_IQ_NO_UNSTAGED_CHANGES', dt("No unstaged changes, and not on a branch; Drush cannot reset. If you have commits that you would like to get rid of, try:\n  git reset HEAD~1\nReplace '1' with the number of commits you have made. Use 'git log' and 'git status' for information."));
    }
    if (!$delete) {
      return drush_set_error('DRUSH_WILL_NOT_DELETE', dt("Not on a branch; Drush will not delete your unstaged changes unless you use the --hard flag."));
    }
    $action_message = dt("Your unstaged changes will be permanently deleted.");
  }
  elseif ($unstaged_changes) {
    $action_message = $delete ? dt("Your working branch !branch will be deleted, along with all unstaged changes.", array('!merges_with' => $branch_merges_with, '!branch' => $branch)) : dt("Your work will be preserved; you can return to it by typing:\n    git checkout !branch\nYour unstaged changes will be re-applied on top of the branch !merges_with", array('!merges_with' => $branch_merges_with, '!branch' => $branch));
  }
  else {
    $action_message = $delete ? dt("Your working branch !branch will be deleted.", array('!merges_with' => $branch_merges_with, '!branch' => $branch)) : dt("Your work will be preserved; you can return to it by typing:\n    git checkout !branch", array('!merges_with' => $branch_merges_with, '!branch' => $branch));
  }
  drush_print($reset_message . $action_message);
  $confirm = drush_confirm(dt("Is it okay to continue?"));
  if (!$confirm) {
    return drush_user_abort();
  }

  if ($delete) {
    drush_shell_exec_interactive("git reset --hard HEAD");
  }
  if (($branch_merges_with != $branch) && ($branch != "(no branch)")) {
    drush_op('chdir', $dir);
    $result = drush_shell_exec_interactive("git checkout %s", $branch_merges_with);
    if ($delete && ($branch != '(no branch)')) {
      $result = drush_shell_exec_interactive("git branch -D %s", $branch);
    }
  }
  drush_op('chdir', $cwd);
  return $issue_info;
}

/**
 * Merge the current branch in with its upstream branch.
 */
function drush_iq_lint() {
  $cwd = getcwd();
  $result = drush_shell_exec('git rev-parse --git-dir 2> /dev/null');
  if (!$result) {
    return drush_set_error('DRUSH_IQ_LINT_NO_GIT', dt("Drush iq-lint only works for files under git revision control"));
  }
  $git_rev_output = drush_shell_exec_output();
  $dir = dirname($git_rev_output[0]);
  drush_op('chdir', $dir);
  $changed_files = _drush_iq_unstaged_changes($dir);
  if (empty($changed_files)) {
    $changed_files = _drush_iq_modified_files_on_branch($dir);
  }
  list($ok, $errors, $not_scanned) = _drush_iq_lint($changed_files);
  drush_op('chdir', $cwd);
  if (!empty($errors)) {
    return drush_set_error('DRUSH_IQ_LINT_FAILED', dt("%count files failed lint check.", array("%count" => count($errors))));
  }
  if (!empty($ok)) {
    drush_log(dt("%count files are ok.", array("%count" => count($ok))), 'ok');
    return TRUE;
  }
  if (!empty($not_scanned)) {
    drush_log(dt("None of the %count changed files are php files.", array("%count" => count($not_scanned))), 'ok');
    return TRUE;
  }
  drush_log(dt("No changed files", 'ok'));
  return TRUE;
}

function _drush_iq_unstaged_changes($dir) {
  $result = drush_shell_cd_and_exec($dir, "git ls-files -m");
  if (!$result) {
    return FALSE;
  }
  return drush_shell_exec_output();
}

function _drush_iq_modified_files_on_branch($dir) {
  $issue_branch = _drush_iq_get_branch_at_dir($dir);
  $branch_merges_with = _drush_iq_get_branch_merges_with($dir);
  $result = drush_shell_cd_and_exec($dir, "git diff --name-only %s..%s", $branch_merges_with, $issue_branch);
  if (!$result) {
    return FALSE;
  }
  return drush_shell_exec_output();
}

function _drush_iq_lint($changed_files) {
  $ok = array();
  $errors = array();
  $not_scanned = array();
  foreach ($changed_files as $changed_file) {
    if ( (substr($changed_file, -4) == '.php') ||
         (substr($changed_file, -4) == '.inc') ||
         (substr($changed_file, -4) == '.module')
      ) {
      $result = drush_shell_exec("php -l %s", $changed_file);
      drush_print(implode("\n", drush_shell_exec_output()));
      if (!$result) {
        $errors[] = $changed_file;
      }
      else {
        $ok[] = $changed_file;
      }
    }
    else {
      $not_scanned[] = $changed_file;
    }
  }
  return array($ok, $errors, $not_scanned);
}

/**
 * Merge the current branch in with its upstream branch.
 */
function drush_iq_merge($number = NULL) {
  $issue_info = array();
  $cwd = getcwd();
  $issue_info = _drush_iq_find_project_dir($number);
  if ($issue_info == FALSE) {
    return FALSE;
  }
  $dir = _drush_iq_project_dir($issue_info);
  $result = drush_shell_cd_and_exec($dir, "git status -s");
  $status_output = drush_shell_exec_output();
  $unstaged_changes = (!empty($status_output));
  $branch = _drush_iq_get_branch_at_dir($dir);
  if (empty($issue_info)) {
    $issue_info = drush_iq_get_info($branch);
  }
  $branch_merges_with = _drush_iq_get_branch_merges_with($dir);
  if ($branch_merges_with === FALSE) {
    return FALSE;
  }
  if (empty($branch_merges_with)) {
    $branch_merges_with = 'master';
  }
  $issue_info['branch'] = $branch;
  $issue_info['mergeBranch'] = $branch_merges_with;

  if (($branch == '(no branch)') || ($branch_merges_with == $branch)) {
    return drush_set_error('DRUSH_IQ_CANNOT_MERGE', dt("Not on a branch; cannot merge."));
  }
  drush_op('chdir', $dir);
  $changed_files = _drush_iq_unstaged_changes($dir);
  if (empty($changed_files)) {
    $changed_files = _drush_iq_modified_files_on_branch($dir);
  }
  list($ok, $errors, $not_scanned) = _drush_iq_lint($changed_files);
  if (!empty($errors)) {
    return drush_set_error('DRUSH_IQ_LINT_FAILED', dt("%count files did not pass lint check; refusing to mereg.", array('%count' => count($errors))));
  }
  if ($unstaged_changes) {
    if (!drush_get_option('no-commit', FALSE)) {
      $commit_comment = _drush_iq_create_commit_comment($issue_info, drush_get_option('committer', TRUE), drush_get_option('message', FALSE));
      // TODO: can we easily ignore *.patch, etc., or easily add only modified files?
      $result = drush_shell_exec("git add -u .");
      $result = drush_shell_exec("git commit -m %s", $commit_comment);
      if (!$result) {
        return drush_set_error('DRUSH_IQ_DIFF_CANNOT_COMMIT', dt("Commit failed."));
      }
    }
    else {
      return drush_set_error('DRUSH_IQ_CANNOT_MERGE', dt("Please commit your unstaged changes before merging, or run command again without the --no-commit option."));
    }
  }
  $confirm = drush_confirm(dt("Would you like to merge the branch !branch with its upstream branch !merges_with? ", array('!merges_with' => $branch_merges_with, '!branch' => $branch)));
  if (!$confirm) {
    return drush_user_abort();
  }
  $squash = drush_get_option('squash', FALSE) ? '--squash' : '';

  $result = drush_shell_exec_interactive("git checkout %s", $branch_merges_with);
  $result = drush_shell_exec_interactive("git merge %s %s", $squash, $branch);
  $result = drush_shell_exec_interactive("git branch -D %s", $branch);
  drush_op('chdir', $cwd);
  return $issue_info;
}

function _drush_iq_create_branch(&$issue_info, $ignore = array()) {
  $project_dir = _drush_iq_project_dir($issue_info);
  $branch_merges_with = _drush_iq_get_branch_merges_with($project_dir);
  if (!$branch_merges_with) {
    $branch_merges_with = _drush_iq_get_branch_at_dir($project_dir);
    if (empty($branch_merges_with)) {
      $result = drush_shell_cd_and_exec($project_dir, "git init");
      $result = drush_shell_cd_and_exec($project_dir, "git add -u .");
      foreach ($ignore as $filename) {
        if (dirname($filename) == $project_dir) {
          $result = drush_shell_cd_and_exec($project_dir, "git rm --cached -- %s", basename($filename));
        }
      }
      $result = drush_shell_cd_and_exec($project_dir, "git commit -m 'Git repository created by Drush iq commands.'");
    }
  }
  else {
    $branch_merges_with = 'origin/' . $branch_merges_with;
  }
  $set_branch_merge_with = FALSE;
  if ($branch_merges_with == "(no branch)") {
    $branch_merges_with = drush_get_option('base', FALSE);
    if (!$branch_merges_with) {
      return drush_set_error('DRUSH_IQ_BRANCH_DETACHED_HEAD', dt("Could not create branch !branch from a detached HEAD state.  It is recommended that you work off of a development branch, such as 7.x-1.x, instead.  If you would like to work from a specific tag, first run `git checkout -b base`, or run this command again with the --base=base option to create a base branch to work from.", array('!branch' => $branchlabel)));
    }
    $result = drush_shell_exec_interactive("git checkout -b %s", $branch_merges_with);
    $set_branch_merge_with = TRUE;
  }
  $issue_info['mergeBranch'] = $branch_merges_with;
  drush_log(dt("Starting at branch !branchlabel", array('!branchlabel' => $branch_merges_with)), 'notice');

  // Make a branch via 'git checkout -b [description]-[issue]'
  $description = $issue_info['title'];
  $comment_number = array_key_exists('patchIndex', $issue_info) ? $issue_info['patchIndex'] : (1 + count($issue_info['comments']));
  $branchlabel = 'drush-iq-' . _drush_iq_make_description($description) . '-' . $issue_info['id'] . '-' . $comment_number;
  $issue_info['branch'] = $branchlabel;
  drush_log(dt("Switching to branch !branchlabel", array('!branchlabel' => $branchlabel)), 'ok');
  $result = drush_shell_exec_interactive("git checkout -b %s %s", $branchlabel, $branch_merges_with);
  if (!$result) {
    return drush_set_error('DRUSH_IQ_BRANCH_CREATE_ERROR', dt("Could not create branch !branch", array('!branch' => $branchlabel)));
  }
  if ($set_branch_merge_with) {
    $result = drush_shell_exec_interactive("git config branch.'%s'.merge %s", $branchlabel, $branch_merges_with);
  }
  return $branchlabel;
}

/**
 * iq-apply-patch command callback
 *
 * Given an issue number, find the most recent patch file
 * attached to it
 */
function drush_iq_apply_patch($number = NULL) {
  if (!isset($number)) {
    $issue_info = drush_iq_get_info_from_browser_window();
  }
  else {
    $issue_info = drush_iq_get_info($number);
  }
  $do_git_operations = !drush_get_option('no-git', FALSE);
  $dry_run = drush_get_option('dry-run', FALSE);
  $result = FALSE;
  if (!$issue_info) {
    return FALSE;
  }
  $project_dir = _drush_iq_project_dir($issue_info);
  if (!$project_dir) {
    return drush_set_error('DRUSH_IQ_APPLY_PATCH_NO_PROJECT', dt('You are not working in a site, and the current directory is not named after the project.  Please cd to the project !project and try again.', array('!project' => $issue_info['projectName'])));
  }
  $issue_info = _drush_iq_select_patch($issue_info, drush_get_option('select', FALSE));
  if (!$issue_info) {
    return FALSE;
  }
  if (!array_key_exists('patchId', $issue_info)) {
    return drush_set_error('DRUSH_NO_PATCHES', dt("Could not find a suitable patch in !issue", array('!issue' => _drush_iq_create_commit_comment($issue_info))));
  }
  $patch_name = $issue_info['patchName'];
  $patch_id = $issue_info['patchId'];
  $patch = $issue_info['patchUrl'];
  drush_log(dt("Drush iq-apply-patch !id: downloading patchfile !patch for project !project", array('!patch' => $patch_name, '!id' => $patch_id, '!project' => $issue_info['projectName'])), 'ok');
  // n.b. the file returned by drush_download_file is either in the cache,
  // or is registered for deletion
  $filename = drush_download_file($patch);
  if (!file_exists($filename)) {
    return drush_set_error('DRUSH_PATCH_NOT_DOWNLOADED', dt("Patch could not be downloaded."));
  }
  $filename = realpath($filename);
  $handle = @fopen($filename, "r");
  if (!$handle) {
    return drush_set_error('DRUSH_PATCH_NOT_READABLE', dt("Patch could not be read."));
  }

  $keep_patch = FALSE;
  if (drush_get_option('keep-patch', FALSE)) {
    $keep_patch = dirname($filename) . '/iq-' . basename($filename);
    copy($filename, $keep_patch);
  }

  $cwd = getcwd();
  drush_op('chdir', $project_dir);

  $try_patch_tool = TRUE;
  if ($do_git_operations && !$dry_run) {
    $result = _drush_iq_create_branch($issue_info, array($filename, $keep_patch));
    if ($result === FALSE) {
      return FALSE;
    }
  }

  $files_to_add = array();
  if ($do_git_operations) {
    // first let us test if we need to add files to the commit later
    $result = drush_shell_exec("git apply -v --summary %s", $filename);
    $git_apply_summary_output = drush_shell_exec_output();
    foreach ($git_apply_summary_output as $line) {
      $words = explode(" ", $line);
      if ($words[1] == 'create' || $words[1] == 'rename') {
        $prefix = substr($words[2], 0, strpos($words[2], '{'));
        $files_to_add[] = $prefix . trim($words[4], '}');
      }
    }
    $other_options = $dry_run ? '--summary' : '';
    $result = drush_shell_exec_interactive("git apply -v $other_options %s", $filename);
    if ($result === FALSE) {
      drush_log(dt("git apply failed; falling back to 'patch' tool"), 'warning');
    }
    else {
      $try_patch_tool = FALSE;
    }
  }

  if ($try_patch_tool) {
    // Try -Np1 first, then -Np0, unless the user selected --no-prefix,
    // in which case we'll try -Np0 followed by -Np1 if that does not work.
    $strip_count = 1;
    if (drush_get_option('no-prefix', FALSE)) {
      $strip_count = 0;
    }

    $result = drush_shell_exec('patch -Np%d --dry-run --batch -d %s -i %s', $strip_count, $project_dir, $filename);
    if (!$result) {
      $strip_count = !$strip_count;
      $result = drush_shell_exec('patch -Np%d --dry-run --batch -d %s -i %s', $strip_count, $project_dir, $filename);
      if (!$result) {
        return drush_set_error('DRUSH_IQ_PATCH_DID_NOT_APPLY', dt("Could not apply the patch with either -Np0 or -Np1; perhaps the patch was rolled for a different version of the project."));
      }
    }
    $dry_run_output = drush_shell_exec_output();
    // Before we apply the patch, run through the dry-run
    // output and make a note of new files that were added
    // by this patch.
    foreach ($dry_run_output as $line) {
      if (substr($line, 0, 14) == 'patching file ') {
        $file = substr($line, 14);
        if (!file_exists($project_dir . '/' . $file)) {
          $files_to_add[] = $file;
        }
      }
    }
    if ($dry_run) {
      drush_print(implode("\n", $dry_run_output));
    }
    else {
      $result = drush_shell_exec_interactive('patch -Np%d --batch --reject-file=- --prefix=/tmp/ -d %s -i %s', $strip_count, $project_dir, $filename);
    }
  }
  if ($do_git_operations && !$dry_run && !drush_get_option('no-commit', FALSE)) {
    $commit_comment = _drush_iq_create_commit_comment($issue_info, drush_get_option('committer', FALSE), drush_get_option('message', FALSE));
    // Add the modified files
    $result = drush_shell_exec("git add -u .");
    // Add the new files
    foreach ($files_to_add as $file) {
      $result = drush_shell_exec("git add %s", $file);
    }
    if (dirname($filename) == $project_dir) {
      $result = drush_shell_exec("git rm --cached -- %s", basename($filename));
      if ($keep_patch) {
        $result = drush_shell_exec("git rm --cached -- %s", basename($keep_patch));
      }
    }
    if (array_key_exists('patchAuthorCredit', $issue_info)) {
      $result = drush_shell_exec("git commit --author=%s -m %s", $issue_info['patchAuthorCredit'], $commit_comment);
    }
    else {
      $result = drush_shell_exec("git commit -m %s", $commit_comment);
    }
  }
  drush_op('chdir', $cwd);
  $changed_files = _drush_iq_unstaged_changes($project_dir);
  if (empty($changed_files)) {
    $changed_files = _drush_iq_modified_files_on_branch($project_dir);
  }
  list($ok, $errors, $not_scanned) = _drush_iq_lint($changed_files);
  if (!empty($errors)) {
    drush_log(dt("%count files did not pass lint check; it would be advisable to use 'drush iq-reset' to discard this patch.", array('%count' => count($errors))), 'warning');
  }
  return $issue_info;
}

function _drush_iq_select_patch($issue_info, $prompt = FALSE) {
  $patch_list = _drush_iq_get_patch_list($issue_info);
  $patches = _drush_iq_collate_patches($issue_info, $patch_list);

  if (!empty($patches)) {
    // If --select, then prompt the user
    if ($prompt) {
      $choices = array();

      foreach ($patch_list as $patch => $info) {
        $index = $info['index'] . $info['suffix'];
        $choices[$patch] = array("$index", ":", $patch);
      }
      $patch = drush_choice($choices, dt("Select a patch to apply:"));
      if ($patch === FALSE) {
        return drush_user_abort();
      }
    }
    // If the issue specification included a comment number, e.g. #1078108-1, then select the patch attached to the specified comment
    elseif (array_key_exists('comment-number', $issue_info)) {
      $index = $issue_info['comment-number'];
      if (!array_key_exists($index, $patches)) {
        return drush_set_error('DRUSH_IQ_NO_PATCH', dt("Error: comment #!index does not exist or does not have a patch.", array('!index' => $index)));
      }
      $patch = $patches[$index];
    }
    else {
      // TODO: See http://drupal.org/node/1078108#comment-5659936 and
      // http://drupal.org/node/1078108#comment-5663932 for additional
      // heuristics we might apply here.
      $patch = array_pop($patches);
    }

    $patch_name = basename($patch);
    $patch_id = $issue_info['id'];
    if (array_key_exists($patch, $patch_list)) {
      $issue_info['patchUrl'] = $patch;
      $issue_info['patchIndex'] = $patch_list[$patch]['index'] . $patch_list[$patch]['suffix'];
      $patch_id .= '-' . $issue_info['patchIndex'];
      if ($patch_list[$patch]['id'] == 0) {
        $issue_info['patchAuthorId'] = $issue_info['authorId'];
        $issue_info['patchAuthorCredit'] = $issue_info['contributors'][$issue_info['authorId']]['authorCredit'];
      }
      else {
        $issue_info['patchAuthorId'] = $issue_info['comments'][$patch_list[$patch]['id']]['contributorId'];
        $issue_info['patchAuthorCredit'] = $issue_info['contributors'][$issue_info['patchAuthorId']]['authorCredit'];
      }
    }
    $issue_info['patchName'] = $patch_name;
    $issue_info['patchId'] = $patch_id;
  }
  return $issue_info;
}

/**
 * This lookup table is used to map items from
 * the issue queue json to the human-readable
 * labels used in the output of iq-info.
 */
function _drush_iq_label_map() {
  return  array(
    'Title' => 'title',
    'ID' => 'id',
    'URL' => 'url',
    'Project' => 'projectTitle',
    'Project URL' => 'projectUrl',
    'Version' => 'version',
    'Component' => 'component',
    'Category' => 'category',
    'Priority' => 'priority',
    'Assigned' => 'assignedName',
    'Status' => 'status',
    'Patch URL' => 'patchUrl',
    'Patch Index' => 'patchIndex',
    'Patch ID' => 'patchId',
  );
}

/**
 * Create a commit comment
 *
 * @param $issue_info array describing an issue; @see drush_iq_get_info()
 *
 * @returns string "#id by contributor1, contributor2: issue title"
 */
function _drush_iq_create_commit_comment($issue_info, $committer = FALSE, $message = FALSE) {
  $contributors = array();

  // If this function is being called because a patch is being
  // created right now, then add the committer to the head of the
  // credits list.
  if ($committer !== FALSE) {
    // If 'TRUE' is passed for the committer, then look up
    // the current user name from git config --list
    if ($committer === TRUE) {
      drush_shell_exec("git config --list | grep '^user.name=' | sed -e 's|[^=]*=||'");
      $output = drush_shell_exec_output();
      $committer = array_pop($output);
    }

    if (!empty($committer)) {
      $contributors[$committer] = $committer;
    }
  }
  // Gather up commit credits, listing most recent contributors first.
  // Everyone who added an attachment gets credit.
  if (!empty($issue_info)) {
    foreach (array_reverse($issue_info['attachments']) as $comment_number => $attachment) {
      if (array_key_exists($attachment['contributorId'], $issue_info['contributors'])) {
        $contributor = $issue_info['contributors'][$attachment['contributorId']]['name'];
        if (!in_array($contributor, $contributors)) {
          $contributors[$contributor] = $contributor;
        }
      }
    }
  }
  $credits = "";
  $credits_prequel = "";
  if (!empty($contributors)) {
    $credits_prequel = " by";
    $credits = " " . implode(', ', $contributors);
  }

  $prequel = "";
  if ($issue_info) {
    $issue_number = $issue_info['id'];
    $issue_title = $issue_info['title'];
    if (!$message) {
      $message = $issue_title;
    }
    $prequel = "Issue #$issue_number$credits_prequel";
  }
  if (!$message) {
    $message = dt("Generated with Drush iq");
  }
  return "$prequel$credits: $message";
}

function drush_iq_get_info_from_browser_window() {
  // Collect an array of potential issue numbers
  $candidate_issues = array();
  $output = array();
  // Linux: use `wmctrl -l` to get a list of all window titles
  $result = drush_shell_exec('wmctrl -l');
  if ($result) {
    $output = drush_shell_exec_output();
  }
  else {
    // Windows: use `tasklist.exe /v`
    $result = drush_shell_exec('tasklist.exe /v');
    if ($result) {
      $output = drush_shell_exec_output();
    }
  }
  // Process Linux or Windows output the same way.
  foreach ($output as $line) {
    if (preg_match('/^[^#]*#([0-9]+):.*drupal.org/', $line, $matches)) {
      $candidate_issues[] = $matches[1];
    }
  }
  // Macintosh: see http://drupal.org/node/1864142

  // Next, iterate through the candidate issues and
  // get info on each one.
  $issue_choices = array();
  $user_selections = array();
  foreach ($candidate_issues as $number) {
    $issue_info = drush_iq_get_info($number);
    // TODO: Screen issues (e.g. check to see if
    // the issue project == the project at cwd,
    // filter out issues for projects that are not enabled in the current site, etc.)
    if ($issue_info) {
      $issue_choices[$number] = $issue_info;
      $user_selections[$number] = '#' . $issue_info['id'] . ' - ' . $issue_info['projectName'] . ': ' . $issue_info['title'];
    }
  }
  if (empty($issue_choices)) {
    return drush_set_error('DRUSH_IQ_NO_ISSUES_FOUND_IN_BROWSER', dt("No issues found in any open browser window."));
  }
  if (count($issue_choices) == 1) {
    return reset($issue_choices);
  }
  $selection = drush_choice($user_selections, dt("Select an issue:"));
  if (!$selection) {
    return drush_user_abort();
  }
  return $issue_choices[$selection];
}

/**
 * Get information about an issue
 *
 * @param $number integer containing the issue number, or string beginning with a "#" and the issue number
 *
 * @returns array
 *  - title             Title of the issue
 *  - id                Issue number
 *  - url               URL to issue page on drupal.org
 *  - projectTitle      Title of the project issue belongs to (e.g. Drush)
 *  - projectName       Name of the project (e.g. drush) - http://drupal.org/project/{projectName}
 *  - projectId
 *  - projectUrl
 *  - version           Project version
 *  - versionId
 *  - authorName       Name of the user who submitted the issue
 *  - authorId
 *  - authorUrl
 *  - assignedName     Name of the user the issue is assigned to
 *  - assignedId
 *  - assignedUrl
 *  - component         Code, documentation, etc.
 *  - category          Bug, feature request, etc.
 *  - priority          minor, normal, major, critical
 *  - priorityId
 *  - status            active, needs work, etc.
 *  - statusId
 *  - created
 *  - changed
 *  - comments          A list (array with numeric keys) of URLs
 *  - attachments       A list (array with numeric keys) of attachments
 *      Array
 *        - contributorId uid of user submitting the patch
 *        - urls        A list of strings pointing to the attachments
 *  - contributors      An associative array keyed by uid of contributors
 *      Array
 *        - name        Name of contributor
 *        - uid         uid of contributor (same as key for this item)
 *        - profile     url to user profile on drupal.org
 */
function drush_iq_get_info($issue_spec) {
  $issue_info_id = drush_iq_issue_number($issue_spec);
  if (empty($issue_info_id)) {
    return drush_set_error('DRUSH_ISSUE_NOT_FOUND', dt("Could not find the issue !issue", array('!issue' => $issue_spec)));
  }
  $issue_info = drush_iq_download_info($issue_info_id);
  return $issue_info;
}

/**
 *  Given an issue specification,
 */
function drush_iq_issue_number($issue_spec) {
  $issue_info_id = array();
  $number = FALSE;
  $issue_site_domain = drush_get_option('issue-site', 'drupal.org');
  $comment_number = FALSE;
  // #1234
  if (substr($issue_spec, 0, 1) == '#') {
    $issue_info_id['id'] = substr($issue_spec, 1);
  }
  // http://drupal.org/node/1234
  elseif (preg_match("#^http://([^/]*)/node/([0-9]*)/*#", $issue_spec, $matches, PREG_OFFSET_CAPTURE)) {
    $issue_site_domain = $matches[1][0];
    $issue_info_id['id'] = $matches[2][0];
  }
  // 1234
  elseif (is_numeric($issue_spec)) {
    $issue_info_id['id'] = $issue_spec;
  }
  // description-of-issue-1234 or description-of-issue-1234-8 (description-issue or description-issue-comment)
  elseif (strpos($issue_spec, ' ') === FALSE) {
    if (preg_match('/-*([0-9]+)(-*(#*[0-9]*[a-z]*))$/', $issue_spec, $matches)) {
      $issue_info_id['id'] = $matches[1];
      if (!empty($matches[3])) {
        $issue_info_id['comment-number'] = $matches[3];
      }
    }
  }
  return $issue_info_id;
}

/**
 *  Given a node id, fetch and decode the issue info json from drupal.org.
 */
function drush_iq_download_info($issue_info_id) {
  $number = $issue_info_id['id'];
  $comment_number = array_key_exists('comment-number', $issue_info_id) ? $issue_info_id['comment-number'] : FALSE;
  $issue_site_domain = drush_get_option('issue-site', 'drupal.org');
  $url = "http://$issue_site_domain/node/$number";
  // Get the json data from d.o.
  $filename = drush_find_tmp() . '/project-issue-' . $number . '.json';
  $filename = _drush_download_file($url . "/project-issue/json", $filename, TRUE);
  if (!empty($filename)) {
    $data = file_get_contents($filename);
  }
  if (!empty($data)) {
    $issue_info = json_decode($data, TRUE);
    // Fixup: if d.o claims that attachment #0 has a contributor, change it to contributor-id
    if (array_key_exists(0, $issue_info['attachments']) && array_key_exists('contributor', $issue_info['attachments'][0]) && !array_key_exists('contributorId', $issue_info['attachments'][0])) {
      $issue_info['attachments'][0]['contributorId'] = $issue_info['attachments'][0]['contributor'];
    }
    // Copy the comment number (in 'subject') from the comment structure
    // to the attachments structure.
    foreach ($issue_info['comments'] as $commentId => $comment) {
      if (array_key_exists($commentId, $issue_info['attachments'])) {
        if (!array_key_exists('commentNumber', $issue_info['attachments'][$commentId])) {
          $issue_info['attachments'][$commentId]['commentNumber'] = $issue_info['comments'][$commentId]['subject'];
        }
      }
    }
    // Fixup: If there is no contributor for attachment #0, then add one in for the issue author
    if (array_key_exists(0, $issue_info['attachments']) && !array_key_exists($issue_info['attachments'][0]['contributorId'], $issue_info['contributors'])) {
      $id = $issue_info['authorId'];
      $issue_info['contributors'][$id] = array(
        'id' => $id,
        'name' => $issue_info['authorName'],
        'url' => $issue_info['authorUrl'],
        );
    }
    // Add in the author credit info
    foreach ($issue_info['contributors'] as $uid => $info) {
      if ($info['name'] != 'System Message') {
        $issue_info['contributors'][$uid]['authorCredit'] = sprintf("%s <%s@%s.no-reply.drupal.org>", $info['name'], $info['name'], $uid);
      }
    }
    if (!empty($issue_info)) {
      if ($comment_number) {
        $issue_info['comment-number'] = $comment_number;
      }
      return $issue_info;
    }
  }
  return drush_set_error('DRUSH_PM_ISSUE_FAILED', dt('Could not fetch issue data from !url', array('!url' => $url)));
}

function _drush_iq_get_patch_list($issue_info) {
  $patch_list = array();

  foreach ($issue_info['attachments'] as $comment_number => $attachment) {
    $comment_patches = array();
    $index = array_key_exists('commentNumber', $attachment) ? $attachment['commentNumber'] : '0';
    foreach ($attachment['urls'] as $url) {
      if (substr($url, -6) == ".patch" || substr($url, -5) == ".diff") {
        $comment_patches[] = $url;
      }
    }
    if (count($comment_patches) == 1) {
      $patch_list[$comment_patches[0]] = array('index' => $index, 'id' => $comment_number, 'suffix' => '');
    }
    else {
      $label = 'a';
      foreach ($comment_patches as $url) {
        $patch_list[$url] = array('index' => $index, 'id' => $comment_number, 'suffix' => $label);
        $label = chr(ord($label) + 1);
      }
    }
  }
  return $patch_list;
}

function _drush_iq_collate_patches($issue_info, $patch_list) {
  $patches = array();

  foreach ($patch_list as $url => $info) {
    $patches[$info['index'] . $info['suffix']] = $url;
    $patches[$info['id'] . $info['suffix']] = $url;
  }

  return $patches;
}

/**
 * Find the project directory associated with the project
 * the specified issue is associated with.
 */
function _drush_iq_project_dir(&$issue_info) {
  $result = FALSE;

  if (array_key_exists('project-dir', $issue_info)) {
    $result = $issue_info['project-dir'];
  }
  elseif(array_key_exists('projectName', $issue_info)) {
    $project_name = $issue_info['projectName'];

    // If our cwd is at the root of the project, then prefer that project over
    // one in some other location.
    $dir = _drush_iq_git_root_dir_from_cwd();
    if (basename($dir) == $project_name) {
      $result = $dir;
    }
    // TODO: Find drush extensions such as drush_extras, drush_make, drubuntu, etc.
    elseif ($project_name == 'drush') {
      $result = DRUSH_BASE_PATH;
    }
    else {
      $phase = drush_bootstrap_max();
      drush_log("bootstrapped to phase $phase");
      if ($phase >= DRUSH_BOOTSTRAP_DRUPAL_CONFIGURATION) {
        $extension_info = drush_get_extensions();
        // TODO: offer to download the project if it is not found?
        if (array_key_exists($project_name, $extension_info)) {
          $result = drush_get_context('DRUSH_DRUPAL_ROOT', '') . '/' . dirname($extension_info[$project_name]->filename);
        }
      }
      if ($phase >= DRUSH_BOOTSTRAP_DRUPAL_ROOT) {
        $root = drush_get_context('DRUSH_DRUPAL_ROOT', FALSE);
        if ($root) {
          if ($project_name == 'drupal') {
            $result = $root;
          }
          else {
            foreach (array('modules', 'sites/all/modules', 'sites/default/modules') as $loc) {
              $path = $root . '/' . $loc . '/' . $project_name;
              if (is_dir($path)) {
                $result = $path;
              }
            }
          }
        }
      }
    }
    if ($result) {
      $issue_info['project-dir'] = $result;
    }
    else {
      drush_log(dt('Could not find the project directory under the bootstrapped site'));
    }
  }

  return $result;
}

function _drush_iq_get_branch(&$issue_info) {
  $branch = FALSE;
  if (array_key_exists('branch', $issue_info)) {
    $branch = $issue_info['branch'];
  }
  else {
    $project_dir = _drush_iq_project_dir($issue_info);
    if ($project_dir) {
      $branch = _drush_iq_get_branch_at_dir($project_dir);
    }
  }
  $issue_info['branch'] = $branch;
  return $branch;
}

function _drush_iq_get_branch_at_dir($dir) {
  $result = drush_shell_cd_and_exec($dir, "git branch");
  $branch_output = drush_shell_exec_output();

  // Return the last non-empty line
  $branch = FALSE;
  while (($branch === FALSE) && !empty($branch_output)) {
    $line = array_shift($branch_output);
    if (!empty($line) && ($line[0] == '*')) {
      $branch_components = explode(' ', $line, 2);
      $branch = $branch_components[1];
    }
  }
  return $branch;
}

function _drush_iq_get_merge_branch_with_remote($merges_with) {
  $remote = _drush_iq_get_best_remote($merges_with);
  return _drush_iq_merge_branch_with_remote($remote, $merges_with);
}

function _drush_iq_merge_branch_with_remote($remote, $merges_with) {
  if (empty($remote)) {
    return $merges_with;
  }
  else {
    return $remote . '/' . $merges_with;
  }
}

function _drush_iq_get_best_remote($merges_with) {
  $remotes = _drush_iq_get_remotes($merges_with);
  if (empty($remotes)) {
    return "";
  }
  elseif (count($remotes) == 1) {
    return $remotes[0];
  }
  elseif (in_array('origin', $remotes)) {
    return 'origin';
  }
  return $remotes[0];
}

function _drush_iq_get_remotes($merges_with) {
  // Check the output of `git branch -r`; we are looking
  // for output "remote/branchname"
  $result = drush_shell_exec("git branch -r");
  $git_branch_output = drush_shell_exec_output();

  $remotes = array();
  foreach($git_branch_output as $line) {
    $line = trim($line);
    if (preg_match("#([^/]+)/(.*)#", $line, $matches)) {
      if ($matches[2] == $merges_with) {
        $remotes[] = $matches[1];
      }
    }
  }
  return $remotes;
}

function _drush_iq_get_branch_merges_with($dir, $branch_label = FALSE) {
  $merges_with = "";
  $default_merges_with = FALSE;
  $alternate_remotes = array();

  if ($branch_label === FALSE) {
    $branch_label = _drush_iq_get_branch_at_dir($dir);
  }
  if ($branch_label == "(no branch)") {
    return $merges_with;
  }
  // We expect that the upstream remote has been configured via
  // `git branch --set-upstream drush-iq-x master`
  $result = drush_shell_cd_and_exec($dir, "git config branch.%s.merge", $branch_label);
  $branch_merge_output = drush_shell_exec_output();
  if (!empty($branch_merge_output)) {
    // Convert the branch name into its abbreviated form.
    $result = drush_shell_exec("git rev-parse --abbrev-ref %s", $branch_merge_output[0]);
    $rev_parse_output = drush_shell_exec_output();
    return $rev_parse_output[0];
  }

  // If the upstream remote has not been configured, then we will look
  // at the output of `git remote show origin` and see if we can find
  // a fallback branch to use.
  $result = drush_shell_cd_and_exec($dir, "git remote show origin -n");
  $show_orgin_output = drush_shell_exec_output();

  foreach ($show_orgin_output as $line) {
    $line = trim($line);
    // Find all of the other branches 'master merges with remote master', etc.
    if (preg_match("/([^ ]+).* with remote (.*)/", $line, $matches)) {
      if ($matches[1] == $matches[2]) {
        $alternate_remotes[] = $matches[2];
      }
    }
  }
  // If there is only one remote, that must be the one we merge with!
  if (count($alternate_remotes) == 1) {
    $merges_with = array_shift($alternate_remotes);
  }
  elseif (!empty($alternate_remotes)) {
    // We could figure out which of these branches was the correct
    // one to use by walking the commit log.  See:
    // http://drupal.org/node/1078108#comment-6335376
    $setupstream = "";
    foreach ($alternate_remotes as $alternate) {
      $setupstream .= "\n  " . dt("git branch --set-upstream !branch !merge", array('!branch' => $branch_label, '!merge' => $alternate));
    }
    return drush_set_error('DRUSH_IQ_CANNOT_FIND_MERGE_BRANCH', dt("Could not determine which branch of !alternatives was the correct merge branch.  Use one of the following commands to select the correct one: !setupstream", array('!alternatives' => implode(',', $alternate_remotes), '!setupstream' => $setupstream)));
  }
  return $merges_with;
}

function _drush_iq_make_description($title) {
  $description = preg_replace('/[^a-z._-]/', '', str_replace(' ', '-', strtolower($title)));

  // Clip the description off at the next dash after the 30th position
  if (strlen($description) > 30) {
    $find_dash = strpos($description, '-', 30);
    if ($find_dash !== FALSE) {
      $description = substr($description, 0, $find_dash);
    }
  }

  return $description;
}

/**
 * Determine the location of the git root directory from the
 * location of the user's current working directory.  If the
 * cwd is somewhere inside the git repo, pop up to the top
 * by using git rev-parse.
 */
function _drush_iq_git_root_dir_from_cwd() {
  $dir = drush_get_context('DRUSH_OLDCWD', drush_cwd());
  exec('git rev-parse --show-toplevel 2> /dev/null', $output);
  if (!empty($output)) {
    $dir = $output[0];
  }
  return $dir;
}
